/**
 * @file config.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * The classes in this file are used for describing characteristics about ecg records
 */
#ifndef ECGLIB_UTIL_CONFIG_LJ_2014_06_06
#define ECGLIB_UTIL_CONFIG_LJ_2014_06_06 1

#include <boost/any.hpp>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include <map>

namespace ecglib { 
	enum class Type {String, Double, Int, Uint};

	struct property {
		Type type;
		boost::any value;
		std::string desc;

		property(const Type typ, const boost::any &val) : type(typ), value(val) {
		}

		property(const Type typ, const boost::any &val, const std::string &description) : type(typ), value(val), desc(description) {
		}
	};

	class from_programoptions {
		public:
			from_programoptions(const std::string &prefix, const po::variables_map &vm) : _prefix(prefix), _vm(vm) {
			}	
		public:
			boost::any get(const std::string &nam, const Type typ) const {
				std::string fullnam = _prefix + nam;

				if(typ == Type::String) {
					return _vm[fullnam].as<std::string>();
				} else if(typ == Type::Double) {
					return _vm[fullnam].as<double>();
				} else if(typ == Type::Int) {
					return _vm[fullnam].as<int>();
				} else if(typ == Type::Uint) {
					return _vm[fullnam].as<unsigned int>();
				} else {
					throw std::logic_error("unsupported type");
				}
			}			

		private:
			std::string _prefix;
			po::variables_map _vm;
	};

	class to_programoptions {
		public:
			to_programoptions(const std::string &prefix, const std::string &name) : _opts(name), _prefix(prefix) {

			}	
		public:
			void set(const std::string &nam, const property &prop) {
				std::string fullnam = _prefix + nam;

				if(prop.type == Type::String) {
					boost::shared_ptr<po::option_description> desc(new po::option_description(fullnam.c_str(), po::value<std::string>()->default_value(boost::any_cast<std::string>(prop.value)),prop.desc.c_str()));
					_opts.add(desc);
				} else if(prop.type == Type::Double) {
					boost::shared_ptr<po::option_description> desc(new po::option_description(fullnam.c_str(), po::value<double>()->default_value(boost::any_cast<double>(prop.value)),prop.desc.c_str()));
					_opts.add(desc);
				} else if(prop.type == Type::Int) {
					boost::shared_ptr<po::option_description> desc(new po::option_description(fullnam.c_str(), po::value<int>()->default_value(boost::any_cast<int>(prop.value)),prop.desc.c_str()));
					_opts.add(desc);
				} else if(prop.type == Type::Uint) {
					boost::shared_ptr<po::option_description> desc(new po::option_description(fullnam.c_str(), po::value<unsigned int>()->default_value(boost::any_cast<unsigned int>(prop.value)),prop.desc.c_str()));
					_opts.add(desc);
				} else {
					throw std::logic_error("unsupported type");
				}
			}

			po::options_description opts() const {
				return _opts;
			}

		private:
			po::options_description _opts;
			std::string _prefix;
	};

	class config {
		public:
			typedef property propertytype;

		public:
			config() {
			}

		public:
			template<class T>
			void load(const T &t) {
				for(auto &r : _props) {
					r.second.value = t.get(r.first, r.second.type);
				}
			}

			template<class T>
			void save(T &t) const {
				for(auto const &r : _props) {
					t.set(r.first, r.second);
				}
			}

			void set(const std::string &nam, const Type typ, const boost::any &val) {
				auto it = _props.find(nam);

				if(it == _props.end()) {
					throw std::logic_error("No such key");
				}

				if(it->second.type != typ) {
					throw std::logic_error("Wrong type");
				}

				it->second.value = val;
			}
	
			

			template<class T>
			T get(const std::string &nam) const {
				auto it = _props.find(nam);

				if(it == _props.end()) {
					throw std::logic_error("No such key:"+nam);
				}

				return boost::any_cast<T>(it->second.value);
			}

		protected:
			void add(const char *nam, const property &&prop) {
				_props.insert(std::pair<std::string, property>(nam, prop));
			}

		protected:
			std::map<std::string, property> _props;
	};
}

#endif
